""" Cython versions of low-level functions for computing vector geometry. """

from libc.math cimport (NAN, M_PI, sqrt,
                        sin, cos, tan, asin, acos, atan, atan2,
                        fmin, fmax, fabs)
from cpython cimport bool
cimport cython
from coordstring cimport CoordString

cdef struct Vector2:
    double x
    double y

cdef struct Vector3:
    double x
    double y
    double z

cdef double dot2(Vector2 u, Vector2 v) nogil:
    return u.x*v.x + u.y*v.y

cdef double dot3(Vector3 u, Vector3 v) nogil:
    return u.x*v.x + u.y*v.y + u.z*v.z

cdef double cross2(Vector2 u, Vector2 v) nogil:
    return u.x*v.y - u.y*v.x

cdef Vector3 cross3(Vector3 u, Vector3 v):
    return Vector3(u.y*v.z - u.z*v.y, u.x*v.z - u.z*v.x, u.x*v.y - u.y*v.x)

cdef Vector2 proj2(Vector2 u, Vector2 v):
    cdef double uv_vv
    uv_vv = dot2(u, v) / dot2(v, v)
    return Vector2(uv_vv*v.x, uv_vv*v.y)

cdef double dist2(Vector2 pt0, Vector2 pt1) nogil:
    return sqrt((pt0.x-pt1.x)**2 + (pt0.y-pt1.y)**2)

@cython.cdivision(True)
cdef double dist_sph(Vector2 pt0, Vector2 pt1) nogil:
    """ Returns distance on a sphere of radius 1 """
    cdef double dx = fabs(pt0.x - pt1.x) * M_PI / 180.0
    cdef double dy = fabs(pt0.y - pt1.y) * M_PI / 180.0
    cdef double y0 = pt0.y * M_PI / 180.0
    cdef double y1 = pt1.y * M_PI / 180.0
    cdef double d_ = 0.0

    if (dx > 0.01) or (dy > 0.01):
        # use spherical law of cosines
        d_ = acos(sin(y0)*sin(y1) + cos(y0)*cos(y1)*cos(dx))

    else:
        # use haversine
        d_ = (2 * asin(sqrt(sin(0.5*dy)**2 + cos(y0)*cos(y1)*sin(0.5*dx)**2)))
    return d_

cdef int fsign(double a):
    """ Return the sign of *a* """
    if a == 0.0:
        return 1
    else:
        return int(a/fabs(a))

cdef int bndlat_sph(Vector2 pt0, Vector2 pt1, double *ymin, double *ymax):
    """ Return the bounding latitudes of a great circle segment on a sphere.
    Returns 0 on success and 1 if the segment is degenerate. """
    cdef int s0 = fsign(pt0.y)
    cdef int s1 = fsign(pt1.y)
    cdef double faz, baz

    cdef double dlam = (pt1.x-pt0.x) * M_PI / 180.0
    cdef double phi0 = pt0.y * M_PI / 180.0
    cdef double phi1 = pt1.y * M_PI / 180.0

    if dlam != 0.0:
        faz = atan2(sin(dlam)*cos(phi1),
                    cos(phi0)*sin(phi1) - sin(phi0)*cos(phi1)*cos(dlam))
        baz = atan2(sin(-dlam)*cos(phi0),
                    cos(phi1)*sin(phi0) - sin(phi1)*cos(phi0)*cos(-dlam))
    elif phi0 == phi1:
        return 1

    if dlam == 0.0 or (s0 != s1):
        ymin[0] = fmin(pt0.y, pt1.y)
        ymax[0] = fmax(pt0.y, pt1.y)
    else:
        if s0 == 1: # northern
            ymin[0] = fmin(phi0, phi1) * 180.0 / M_PI

            if ((fabs((faz + M_PI) % (2*M_PI) - M_PI) < 0.5*M_PI) and
                (fabs((baz + M_PI) % (2*M_PI) - M_PI) < 0.5*M_PI)):
                # both azimuths "point up"
                ymax[0] = acos(fabs(sin(faz)*cos(phi0))) * 180.0 / M_PI
            else:
                ymax[0] = fmax(phi0, phi1) * 180.0 / M_PI

        else:
            ymax[0] = fmax(phi0, phi1) * 180.0 / M_PI

            if ((fabs((faz + M_PI) % (2*M_PI) - M_PI) > 0.5*M_PI) and
                (fabs((baz + M_PI) % (2*M_PI) - M_PI) > 0.5*M_PI)):
                # both azimuths "point down"
                ymin[0] = -acos(fabs(sin(faz)*cos(phi0))) * 180.0 / M_PI
            else:
                ymin[0] = fmin(phi0, phi1) * 180.0 / M_PI
    return 0

def bbox(CoordString cs):
    cdef double xmin, xmax, ymin, ymax, x, y, x0, y0
    cdef int i = 1, n = len(cs)
    if n == 0:
        return (NAN, NAN, NAN, NAN)
    x0 = cs.getX(0)
    y0 = cs.getY(0)
    xmin = x0
    xmax = x0
    ymin = y0
    ymax = y0
    while i != n:
        x = cs.getX(i)
        y = cs.getY(i)
        xmin = fmin(xmin, x)
        xmax = fmax(xmax, x)
        ymin = fmin(ymin, y)
        ymax = fmax(ymax, y)
        x0 = x
        y0 = y
        i += 1
    return (xmin, ymin, xmax, ymax)

@cython.cdivision(True)
cdef Vector3 sph2cart(Vector2 a):
    """ Convert a (lambda, phi) coordinate on a sphere with an origin at
    (0, 0, 0) to an (x, y, z) coordinate. """
    cdef double theta = 90.0 - a.y
    return Vector3(sin(M_PI*theta/180.0)*cos(M_PI*a.x/180.0),
                   sin(M_PI*theta/180.0)*sin(M_PI*a.x/180.0),
                   cos(M_PI*theta/180.0))

@cython.cdivision(True)
cdef Vector2 cart2sph(Vector3 a):
    """ Convert an (x, y, z) coordinate to a (lambda, phi) coordinate on a
    sphere with an origin at (0, 0, 0). """
    cdef double lon = 0.0
    cdef double lat = 0.0
    if fabs(a.x) > 1e-8:
        lon = atan2(a.y, a.x)
    else:
        lon = asin(a.y / sqrt(a.x*a.x + a.y*a.y))
    if fabs(a.z) > 1e-8:
        lat = 0.5*M_PI - atan(sqrt(a.x*a.x + a.y*a.y)/a.z)
    else:
        lat = 0.5*M_PI - acos(a.z / sqrt(a.x*a.x + a.y*a.y + a.z*a.z))
    return Vector2(lon*180.0/M_PI, lat*180.0/M_PI)

cdef Vector3 eulerpole_cart(Vector3 a, Vector3 b):
    return cross3(a, b)

cdef Vector3 eulerpole(Vector2 a, Vector2 b):
    cdef Vector3 ac = sph2cart(a)
    cdef Vector3 bc = sph2cart(b)
    cdef ep = cross3(ac, bc)
    return ep

cdef double azimuth(Vector2 pt1, Vector2 pt2):
    """ Return azimuth in radians between two points on a plane. """
    return atan2(pt2.x-pt1.x, pt2.y-pt1.y)

@cython.cdivision(True)
cdef double azimuth_sph(Vector2 pt1, Vector2 pt2):
    """ Return azimuth in radians between two points on a sphere. Inputs taken
    as degrees of (lon, lat). """
    cdef double dlon = (pt2.x - pt1.x) * M_PI / 180.0
    cdef double y1 = pt1.y * M_PI / 180.0
    cdef double y2 = pt2.y * M_PI / 180.0
    return atan2(sin(dlon), cos(y1) * tan(y2) - sin(y1) * cos(dlon))

def length(CoordString cs):
    """ Compute planar length of CoordString """
    cdef int n = len(cs)
    if cs.ring:
        n += 1
    cdef int i = 0
    cdef double d = 0.0
    cdef double x0, y0, x1, y1
    x0 = cs.getX(i)
    y0 = cs.getY(i)
    while i != (n-1):
        x1 = cs.getX(i+1)
        y1 = cs.getY(i+1)
        d += sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0))
        x0 = x1
        y0 = y1
        i += 1
    return d

def pt_nearest_planar(double x, double y,
                      double endpt0_0, double endpt0_1,
                      double endpt1_0, double endpt1_1):
    """ Determines the point on a segment defined by pairs endpt1
    and endpt2 nearest to a point defined by x, y.
    Returns the a nested tuple of (point, distance)
    """

    cdef double u0, u1, v0, v1
    cdef Vector2 u_on_v, u_int, pt, pt0, pt1
    cdef Vector2 u, v

    pt = Vector2(x, y)
    pt0 = Vector2(endpt0_0, endpt0_1)
    pt1 = Vector2(endpt1_0, endpt1_1)

    u = Vector2(x - pt0.x, y - pt0.y)
    v = Vector2(pt1.x - pt0.x, pt1.y - pt0.y)
    u_on_v = proj2(u, v)
    u_int = Vector2(u_on_v.x + pt0.x, u_on_v.y + pt0.y)

    # Determine whether u_int is inside the segment
    # Otherwise return the nearest endpoint
    if fabs(pt0.x - pt1.x) < 1e-4:  # Segment is near vertical
        if u_int.y > fmax(pt0.y, pt1.y):
            if pt0.y > pt1.y:
                return ((pt0.x, pt0.y), dist2(pt0, pt))
            else:
                return ((pt1.x, pt1.y), dist2(pt1, pt))

        elif u_int.y < fmin(pt0.y, pt1.y):
            if pt0.y < pt1.y:
                return ((pt0.x, pt0.y), dist2(pt0, pt))
            else:
                return ((pt1.x, pt1.y), dist2(pt1, pt))

        else:
            return ((u_int.x, u_int.y), dist2(u_int, pt))

    else:

        if u_int.x > fmax(pt0.x, pt1.x):
            if pt0.x > pt1.x:
                return ((pt0.x, pt0.y), dist2(pt0, pt))
            else:
                return ((pt1.x, pt1.y), dist2(pt1, pt))

        elif u_int.x < fmin(pt0.x, pt1.x):
            if pt0.x < pt1.x:
                return ((pt0.x, pt0.y), dist2(pt0, pt))
            else:
                return ((pt1.x, pt1.y), dist2(pt1, pt))

        else:
            return ((u_int.x, u_int.y), dist2(u_int, pt))

# Currently unused, but could be employed to optimize pyproj function calls
ctypedef tuple (*fwd_t)(double, double, double, double)
ctypedef tuple (*inv_t)(double, double, double, double)

cdef double _along_distance(object fwd, object inv, double x0, double y0,
        double xp, double yp, double az, double f):
    """ Return the distance between a point distance f along azimuth az away
    from (x0, y0) from (xp, yp). """
    cdef double trialx, trialy, _, d
    (trialx, trialy, _) = fwd(x0, y0, az, f)
    (_, _, d) = inv(trialx, trialy, xp, yp)
    return d

cdef double _along_distance_gradient(object fwd, object inv, double x0, double y0,
        double xp, double yp, double az, double f, double dx):
    """ Return the numerical gradient in terms of f of _along_distance """
    cdef double d1, d2
    d1 = _along_distance(fwd, inv, x0, y0, xp, yp, az, f)
    d2 = _along_distance(fwd, inv, x0, y0, xp, yp, az, f+dx)
    return (d2-d1)/dx

def pt_nearest_proj(object fwd, object inv, tuple pt, double[:] endpt0, double[:] endpt1,
        float tol=0.1, int maxiter=100):
    """ Given geodetic functions *fwd* and *inv*, a *pt*, and an arc from
    *endpt1* to *endpt2*, return the point on the arc that is nearest *pt*.

    Scheme employs a bisection minimization method. Iteration continues until a
    tolerance *tol* in meters is reached, or *maxiter* iterations are
    exhausted. If the iteration limit is reached, a ConvergenceError is raised.
    """
    cdef double az, az2, L
    (az, az2, L) = inv(endpt0[0], endpt0[1], endpt1[0], endpt1[1])

    # Detect whether the nearest point is at an endpoint
    cdef double grad0, grad1, d
    grad0 = _along_distance_gradient(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, 0, 1e-7*L)
    grad1 = _along_distance_gradient(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, L, 1e-7*L)
    if grad0 > 0:
        d = _along_distance(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, 0)
        return endpt0, d
    elif grad1 < 0:
        d = _along_distance(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, L)
        return endpt1, d

    # Bisection iteration
    cdef double x0 = 0.0, x1 = 1.0, xm = 0.0, xn = 0.0, yn = 0.0
    cdef int i = 0
    cdef double dx = tol + 1.0
    cdef double grad

    while dx > tol:
        if i == maxiter:
            raise ConvergenceError("Maximum iterations exceeded")
        xm = 0.5 * (x0 + x1)
        grad = _along_distance_gradient(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, xm*L, 1e-7*L)
        if grad > 0:
            dx = fabs(x1-xm) * L
            x1 = xm
        else:
            dx = fabs(x0-xm) * L
            x0 = xm
        i += 1

    (xn, yn, _) = fwd(endpt0[0], endpt0[1], az, xm*L)
    d = _along_distance(fwd, inv, endpt0[0], endpt0[1], pt[0], pt[1], az, xm*L)
    return (xn, yn), d

class ConvergenceError(Exception):
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message




import numpy as np
from ..crs import CartesianCRS

def coordinate_transformer(crs1, crs2):
    if crs1 == crs2:
        def transformer(x, y):
            return x, y
    else:
        def transformer(x, y):
            return crs1.transform(crs2, x, y)
    return transformer

class CoordinateGenerator(object):
    """ Class that generates coordinates from grid indices in a specified
    coordinate system. """

    def __init__(self, transform, size, transform_crs=CartesianCRS, output_crs=CartesianCRS):
        """ Creates a CoordinateGenerator. A CoordinateGenerator generates
        world positions (geographical or projected) from indices or slices
        based on a RegularGrid transform tuple.

        Parameters
        ----------
        transform : tuple
            RegularGrid transform tuple in the form (x0, y0, dx, dy, sx, sy)
        size : tuple
            grid size for slicing
        transform_crs : CRS
            coordinate system used by the transform tuple
        output_crs : CRS
            coordinate system of the positions generated by indexing the
            CoordinateGenerator

        Usage
        -----
        # grid is a RegularGrid instance
        cg = grid.coordinates(crs=LonLatWGS84)

        # indexing returns the coordinates of cells in the grid, transformed to
        # the desired coordinate system
        x, y = cg[100, 250]

        # slicing is supported
        X, Y = cg[:, 200:32:-2]
        allX, allY = cg[:,:]
        """
        self.transform = transform
        self.size = size
        self._coordtransformer = coordinate_transformer(transform_crs, output_crs)

    def __getitem__(self, key):
        if len(key) != 2:
            raise KeyError("CoordinateGenerator indices must be length 2")
        ikey, jkey = key
        if isinstance(ikey, slice):
            ikey = np.arange(*ikey.indices(self.size[0]))
        if isinstance(jkey, slice):
            jkey = np.arange(*jkey.indices(self.size[1]))
        if hasattr(ikey, "__len__") and hasattr(jkey, "__len__") and \
                (len(ikey) > 1) and (len(jkey) > 1):
            jkey, ikey = np.meshgrid(jkey, ikey)
        T = self.transform
        x_ = T[0] + jkey*T[2] + ikey*T[4] + 0.5*T[2] + 0.5*T[4]
        y_ = T[1] + ikey*T[3] + jkey*T[5] + 0.5*T[3] + 0.5*T[5]
        return self._coordtransformer(x_, y_)




"""
Data input functions for Karta

Used available drivers to read input data and return Karta geometry objects.
"""
import os
from numbers import Number
from osgeo import ogr
import picogeojson
from . import geometry
from . import _shp as shp
from . import _gpx as gpx
from ..crs import GeographicalCRS, ProjectedCRS, LonLatWGS84
from .. import errors

### GeoInterface functions ###

def from_shape(obj, properties=None):
    """ Read a __geo_interface__ dictionary and return an appropriate karta
    object """
    return _from_shape(obj.__geo_interface__, None)

def _from_shape(d, properties):
    if d is None:
        return None
    elif d["type"] == "Feature":
        p = d["properties"]
        return _from_shape(d["geometry"], p)
    else:
        c = d["coordinates"]
        if c is None:
            return None
        elif d["type"] == "Point":
            return geometry.Point(c, properties=properties)
        elif d["type"] == "LineString":
            return geometry.Line(c, properties=properties)
        elif d["type"] == "Polygon":
            subs = [geometry.Polygon(c[i]) for i in range(1, len(c))]
            return geometry.Polygon(c[0], properties=properties, subs=subs)
        elif d["type"] == "MultiPoint":
            return geometry.Multipoint(c, properties=properties)
        elif d["type"] == "MultiLineString":
            return geometry.Multiline(c, properties=properties)
        elif d["type"] == "MultiPolygon":
            return geometry.Multipolygon(c, properties=properties)
        else:
            raise NotImplementedError("Geometry type {0} not "
                                      "implemented".format(d["type"]))

### GeoJSON functions ###

def read_geojson(f, crs=LonLatWGS84):
    """ Parse GeoJSON and return a list of geometries.

    f : file-like object or str
        file object to read from or a GeoJSON string
    crs : karta.crs.CRS
        CRS object to bind to new geometries
    """
    def _convert_crs(crsdict):
        # Deprecated
        if crsdict.get("type", None) not in ("name", "link"):
            crs = LonLatWGS84
        elif crsdict["type"] == "name":
            crs = picogeojson.GeoJSONNamedCRS(crsdict["properties"]["name"])
        elif crsdict["type"] == "link":
            crs = picogeojson.GeoJSONLinkedCRS(crsdict["properties"]["href"],
                                               crsdict["properties"]["type"])
        return crs

    def convert(geom, **kw):
        if isinstance(geom, picogeojson.Feature):
            res = [convert_feature(geom, **kw)]
        elif isinstance(geom, picogeojson.FeatureCollection):
            res = [convert_feature(f, **kw) for f in geom.features]
        elif isinstance(geom, picogeojson.GeometryCollection):
            res = [convert(item, **kw) for item in geom.geometries]
        else:
            res = convert_geometry(geom, **kw)
        return res

    def convert_geometry(geom, **kw):
        # TODO: Requires clean-up; was written before properties vs data was
        # clearly worked out. Hacks to bring it up to speed are ugly.
        if isinstance(geom, picogeojson.Point):
            kw.setdefault("properties", {})
            kw.setdefault("data", {})
            kw["properties"].update(kw["data"])
            del kw["data"]
            return geometry.Point(geom.coordinates, **kw)
        elif isinstance(geom, picogeojson.LineString):
            kw.setdefault("properties", {})
            kw.setdefault("data", {})
            kw["properties"].update(kw["data"])
            del kw["data"]
            return geometry.Line(geom.coordinates, **kw)
        elif isinstance(geom, picogeojson.Polygon):
            kw.setdefault("properties", {})
            kw.setdefault("data", {})
            kw["properties"].update(kw["data"])
            del kw["data"]
            return geometry.Polygon(geom.coordinates[0],
                                    subs=geom.coordinates[1:],
                                    **kw)
        elif isinstance(geom, picogeojson.MultiPoint):
            return geometry.Multipoint(geom.coordinates, **kw)
        elif isinstance(geom, picogeojson.MultiLineString):
            return geometry.Multiline(geom.coordinates, **kw)
        elif isinstance(geom, picogeojson.MultiPolygon):
            return geometry.Multipolygon(geom.coordinates, **kw)
        else:
            raise TypeError("{0} is a not a GeoJSON entity".format(type(geom)))

    def convert_feature(feat, **kw):
        properties = {}
        data = {}
        if isinstance(feat.geometry, (picogeojson.MultiPoint,
                                      picogeojson.MultiLineString,
                                      picogeojson.MultiPolygon)):
            n = len(feat.geometry.coordinates)
            for k, v in feat.properties.items():
                try:
                    if len(v) == n:
                        data[k] = v
                    else:
                        properties[k] = v
                except TypeError:
                    properties[k] = v
        else:
            properties.update(feat.properties)
        # if len(data) == 0:
        #     data = {}
        # else:
        #     for key, val in data.items():
        #         if any(isinstance(a, Number) or hasattr(a, "dtype") for a in val):
        #             for i in range(len(val)):
        #                 if val[i] is None:
        #                     val[i] = float('nan')
        kw["data"] = data
        kw["properties"] = properties
        return convert(feat.geometry, **kw)

    R = picogeojson.Deserializer()
    geom = R(f)
    return convert(geom, crs=crs)

def _geojson_properties2karta(properties, n):
    """ Takes a dictionary (derived from a GeoJSON properties object) and
    divides it into singleton properties and *n*-degree data. """
    props = {}
    data = {}
    for (key, value) in properties.items():
        if isinstance(value, list) or isinstance(value, tuple):
            if len(value) == n:
                data[key] = value
            else:
                raise ValueError("properties must be singleton or per-vertex")
        else:
            props[key] = value
    return props, data

### Shapefile functions ###

def get_filenames(stem, check=False):
    """ Given a filename basename, return the associated shapefile paths. If
    `check` is True, ensure that the files exist."""
    if stem.endswith(".shp"):
        stem = stem[:-4]
    shp = stem + '.shp'
    shx = stem + '.shx'
    dbf = stem + '.dbf'
    if check:
        for fnm in (shp, shx, dbf):
            if not os.path.isfile(fnm):
                raise Exception('missing {0}'.format(fnm))
    return {'shp':shp, 'shx':shx, 'dbf':dbf}

def ogr_read_shapefile(stem):
    fnms = get_filenames(stem)
    driver = ogr.GetDriverByName("ESRI Shapefile")
    ds = driver.Open(fnms["shp"], 0)
    layer = ds.GetLayer()

    try:
        _geoms = [_from_shape(gi, p)
                 for (gi,p) in zip(shp.ogr_read_geometries(layer),
                                   shp.ogr_read_attributes(layer))]
        crs = ogr_parse_srs(layer)
    finally:
        del ds, driver, layer

    geoms = []
    for g in _geoms:
        if isinstance(g, list):
            for part in g:
                part.crs = crs
            geoms.append(g)
        elif g is not None:
            g.crs = crs
            geoms.append(g)
    return geoms

def ogr_parse_srs(lyr):
    """ Given an OGR type with a `GetSpatialRef` method, return a matching CRS
    object. """
    srs = lyr.GetSpatialRef()
    if srs is None:
        crs = LonLatWGS84
    else:
        name = srs.GetAttrValue('PROJCS')
        if srs.IsGeographic():
            spheroid = "+a={a} +f={f}".format(a=srs.GetSemiMajor(),
                                              f=1.0/srs.GetInvFlattening())
            crs = GeographicalCRS(spheroid, name)
        else:
            crs = ProjectedCRS(srs.ExportToProj4(), name=name)
    return crs

# convenience binding
read_shapefile = ogr_read_shapefile


### GPX functions ###

def read_gpx_waypts(fnm):
    gpx_doc = gpx.GPX(fnm)
    return [_waypt2pt(pt) for pt in gpx_doc.waypts]

def read_gpx_tracks(fnm):
    gpx_doc = gpx.GPX(fnm)
    return [_track2lines(trk) for trk in gpx_doc.tracks]

def _waypt2pt(waypt):
    return geometry.Point(waypt.lonlat,
                          properties=waypt.properties,
                          crs=LonLatWGS84)

def _seg2line(seg):
    return geometry.Line([pt.lonlat for pt in seg.trkpts],
                         properties=seg.properties,
                         crs=LonLatWGS84)

def _track2lines(track):
    return [_seg2line(seg) for seg in track.trksegs]




""" Low-level functions for reading ESRI ASCII grids """

import numpy as np

def aairead(fnm):
    """ Read an existing ASCII grid file and return a Numpy array and a
    dictionary of header information. """
    h = []
    with open(fnm, 'r') as f:
        # Count the number of header entries
        cnt = 0
        while True:
            l = f.readline()
            if l.split(None, 1)[0].lower() in ['nrows', 'ncols',
                    'yllcenter', 'xllcenter', 'yllcorner', 'xllcorner',
                    'cellsize', 'nodata_value']:
                cnt += 1
            else:
                break

        # Read the header, then the array data
        f.seek(0)
        for _ in range(cnt):
            h.append(f.readline())
        data = f.readlines()

    # Store data internally
    hs = [rec.split(None, 1) for rec in h]
    hdr = dict([(rec[0].lower(), float(rec[1])) for rec in hs])
    hdr['ncols'] = int(hdr['ncols'])
    hdr['nrows'] = int(hdr['nrows'])
    if 'yllcenter' not in hdr.keys():
        hdr['yllcenter'] = None

    if 'xllcenter' not in hdr.keys():
        hdr['xllcenter'] = None

    if 'yllcorner' not in hdr.keys():
        hdr['yllcorner'] = None

    if 'xllcorner' not in hdr.keys():
        hdr['xllcorner'] = None

    if 'nodata_value' not in hdr.keys():
        hdr['nodata_value'] = -9999

    check_header(hdr)

    f = lambda l: [float(i) for i in l.split()]
    data_f = list(map(f, data))
    data_a = np.array(data_f)
    data_a[data_a==hdr['nodata_value']] = np.nan

    return data_a, hdr


def check_header(hdr):
    """ Make sure that all required header records are present, as well as both
    centered and corner spatial references. """
    for field in ('ncols', 'nrows', 'cellsize'):
        if field not in hdr:
            raise ValueError("{0} not set in header".format(field.upper()))

    d = hdr['cellsize']

    if hdr.get('yllcenter') is None:
        if hdr.get('yllcorner') is None:
            raise ValueError('YLL reference not set in header')
        else:
            hdr['yllcenter'] = hdr['yllcorner'] + d / 2.0
    else:
        hdr['yllcorner'] = hdr['yllcenter'] - d / 2.0

    if hdr.get('xllcenter') is None:
        if hdr.get('xllcorner') is None:
            raise ValueError('XLL reference not set in header')
        else:
            hdr['xllcenter'] = hdr['xllcorner'] + d / 2.0
    else:
        hdr['xllcorner'] = hdr['xllcenter'] - d / 2.0

    return hdr




from libc.math cimport NAN, M_PI, sin, cos, fmin, fmax, isnan
from cpython cimport bool
from coordstring cimport CoordString
from vectorgeo cimport (Vector2, Vector3,
                        cross2, cross3,
                        azimuth_sph, cart2sph, sph2cart,
                        eulerpole, eulerpole_cart)
import heapq

cdef bool isbetween_inc(double a, double b, double c):
    return fmin(a, c) <= b <= fmax(a, c)

cdef bool isbetween_incl(double a, double b, double c):
    return fmin(a, c) <= b < fmax(a, c)

cdef bool isbetween_incr(double a, double b, double c):
    return fmin(a, c) < b <= fmax(a, c)

ctypedef bool (*isbetween_t)(double, double, double)

def bboxes_overlap(tuple bb0, tuple bb1):
    """ Return whether planar bboxes overlap.
    """
    cdef float dx = 0.0
    cdef float dy = 0.0
    dx = fmin(bb0[2], bb1[2]) - fmax(bb0[0], bb1[0])
    dy = fmin(bb0[3], bb1[3]) - fmax(bb0[1], bb1[1])
    if dx == 0.0:
        dx = 1.0
    elif dx < 0.0:
        dx = 0.0
    if dy ==  0.0:
        dy = 1.0
    elif dy < 0.0:
        dy = 0.0
    if dx*dy == 0.0:
        return False
    else:
        return True

def all_intersections(CoordString a, CoordString b):
    """ Brute-force intersection search with len(a)*len(b) complexity """
    cdef int na = len(a)
    cdef int nb = len(b)
    cdef int i = 0, j = 0
    cdef double x0, x1, x2, x3, y0, y1, y2, y3
    cdef xi, yi
    cdef list intersections = []

    if not a.ring:
        na -= 1
    if not b.ring:
        nb -= 1

    for i in range(na):
        x0 = a.getX(i)
        x1 = a.getX(i+1)
        y0 = a.getY(i)
        y1 = a.getY(i+1)
        for j in range(nb):
            x2 = b.getX(j)
            x3 = b.getX(j+1)
            y2 = b.getY(j)
            y3 = b.getY(j+1)
            xi, yi = intersection(x0, x1, x2, x3, y0, y1, y2, y3)
            if not isnan(xi):
                intersections.append((xi, yi))
    return intersections

cdef class Event(object):
    """ An Event represents an event in a sweep-line algorithm that signals a
    change in the sweep-line datastructure. This event signals the entrance and
    exits of line segments and segment intersections.
    """
    cdef public tuple seg
    cdef public int kind
    cdef public int color
    cdef public tuple left
    cdef public tuple right

    def __init__(self, tuple seg, int kind, int color):
        self.seg = seg
        self.kind = kind
        self.color = color
        self.left = None
        self.right = None

    def __richcmp__(self, Event other, int op):
        if op == 0:
            return self.seg < other.seg
        elif op == 1:
            return self.seg <= other.seg
        elif op == 2:
            return self.seg == other.seg
        elif op == 3:
            return self.seg != other.seg
        elif op == 4:
            return self.seg > other.seg
        elif op == 5:
            return self.seg >= other.seg
        else:
            raise ValueError("no __richcmp__ operation for op == %d" % op)

class EventQueuePlaceholder(object):
    """ This is a placeholder for an event-queue class. """

    def __init__(self):
        self.queue = []

    def __len__(self):
        return len(self.queue)

    def insert(self, double x, Event event):
        heapq.heappush(self.queue, (x, event))
        return

    def takefirst(self):
        """ return a flag and the first item from the queue. if flag == 1, then
        next item in the queue happens at the same time. """
        cdef int flag = 0
        cdef double x
        cdef Event event
        x, event = heapq.heappop(self.queue)
        if (len(self.queue) != 0) and (x == self.queue[0][0]):
            flag = 1
        return flag, x, event

class SweepLinePlaceholder(object):
    """ This is a placeholder for a sweep-line class. """
    def __init__(self, spherical=False):
        self.queue = []
        self.spherical = spherical

    def __iter__(self):
        cdef tuple item
        for item in self.queue:
            yield item

    def _get_slope_height(self, tuple seg, double x):
        cdef double m = 0.0, h = 0.0
        cdef tuple interx
        if self.spherical:
            m = -(azimuth_sph(Vector2(seg[0], seg[1]), Vector2(seg[2], seg[3]))*180/M_PI - 90)
            # compute intersection between seg and a meridional geodesic passing through x
            interx = intersection_sph(seg[0], seg[2], x, x, seg[1], seg[3], -90, 90)
            h = interx[1]
        else:
            if seg[3] == seg[1]:
                return m, seg[1]
            m = (seg[3]-seg[1]) / (seg[2]-seg[0])
            h = m*(x-seg[0]) + seg[1]
        return m, h

    def insert(self, tuple item, double x):
        if len(self.queue) == 0:
            self.queue.append(item)
            return

        cdef double m, y, m_, y_
        cdef int i = 0
        m, y = self._get_slope_height(item, x)
        m_, y_ = self._get_slope_height(self.queue[i], x)
        while (y_ < y) or ((y_ == y) and (m_ < m)):
            i += 1
            if i == len(self.queue):
                break
            m_, y_ = self._get_slope_height(self.queue[i], x)
        self.queue.insert(i, item)
        return

    def leftof(self, tuple item):
        cdef int idx
        idx = self.queue.index(item)
        if idx == 0:
            return None
        else:
            return self.queue[idx-1]

    def rightof(self, tuple item):
        cdef int idx
        idx = self.queue.index(item)
        if idx == len(self.queue) - 1:
            return None
        else:
            return self.queue[idx+1]

    def remove(self, item):
        cdef int idx
        idx = self.queue.index(item)
        self.queue.pop(idx)
        return

def intersects(CoordString a, CoordString b):
    """ Shamos-Hoey intersection detection algorithm """

   # corner cases:
   # - simultaneous events          handled
   # - parallel overlapping lines   handled
   # - triple intersections         handled(?)
   # - vertical lines               handled
   # - self-intersections           handled
   #
   # all of the above need tests to verify

    cdef int na = len(a), nb = len(b)
    if a.ring:
        na = na + 1
    if b.ring:
        nb = nb + 1

    # initialize event queue
    #
    # event kind may be:
    # 0 : left point
    # 1 : right point
    # 2 : bottom point on vertical segment
    # 3 : top point on vertical segment
    # 4 : intersection (unused)
    cdef int i = 0
    cdef object event_queue = EventQueuePlaceholder()
    cdef double x0, x1, y0, y1
    cdef tuple seg

    for i in range(na-1):
        x0 = a.getX(i)
        x1 = a.getX(i+1)
        y0 = a.getY(i)
        y1 = a.getY(i+1)
        seg = (x0, y0, x1, y1, 0)
        if x0 == x1:
            if y0 < y1:
                event_queue.insert(x0, Event(seg, 2, 0))
                event_queue.insert(x0, Event(seg, 3, 0))
            else:
                event_queue.insert(x0, Event(seg, 3, 0))
                event_queue.insert(x0, Event(seg, 2, 0))
        elif x0 < x1:
            event_queue.insert(x0, Event(seg, 0, 0))
            event_queue.insert(x1, Event(seg, 1, 0))
        else:
            event_queue.insert(x1, Event(seg, 0, 0))
            event_queue.insert(x0, Event(seg, 1, 0))

    for i in range(nb-1):
        x0 = b.getX(i)
        x1 = b.getX(i+1)
        y0 = b.getY(i)
        y1 = b.getY(i+1)
        seg = (x0, y0, x1, y1, 1)
        if x0 == x1:
            if y0 < y1:
                event_queue.insert(x0, Event(seg, 2, 1))
                event_queue.insert(x0, Event(seg, 3, 1))
            else:
                event_queue.insert(x0, Event(seg, 3, 1))
                event_queue.insert(x0, Event(seg, 2, 1))
        elif x0 < x1:
            event_queue.insert(x0, Event(seg, 0, 1))
            event_queue.insert(x1, Event(seg, 1, 1))
        else:
            event_queue.insert(x1, Event(seg, 0, 1))
            event_queue.insert(x0, Event(seg, 1, 1))

    # begin sweep search
    cdef double x
    cdef double yA
    cdef int event_type, segcolor
    cdef object sweepline = SweepLinePlaceholder()
    cdef Event event = None, ev = None
    cdef tuple segA, segB
    cdef list current_events = [], neighbours = []

    while len(event_queue) != 0:

        # get all simultaneous events
        current_events = []
        flag = 1
        while flag == 1:
            flag, x, event = event_queue.takefirst()
            current_events.append((x, event))

        # for all segments that begin, insert into sweepline
        for x, event in current_events:
            if event.kind == 0:
                sweepline.insert(event.seg, x)

        # for all segments that end, get the neighbours that are not also
        # ending and remove from sweepline
        for _, event in current_events:
            if event.kind == 1:
                segA = event.seg
                segB = event.seg
                while segA is not None and \
                        segA in (ev.seg for x, ev in current_events if ev.kind==1):
                    segA = sweepline.leftof(segA)
                while segB is not None and \
                        segB in (ev.seg for x, ev in current_events if ev.kind==1):
                    segB = sweepline.rightof(segB)
                event.left = segA
                event.right = segB
                sweepline.remove(event.seg)

        # check for intersections due to last set of events
        for x, event in current_events:

            seg = event.seg

            if (event.kind == 2):
                # vertical segment - check whether span crosses any segments
                for segA in sweepline:

                    if seg[4] == segA[4]:
                        continue

                    yA = (segA[3]-segA[1]) / (segA[2]-segA[0])*(seg[0]-segA[0]) + segA[1]
                    if (seg[1] <= yA) and (seg[3] >= yA):
                        return True

                    if (seg[3] <= yA) and (seg[1] >= yA):
                        return True

                # can skip kind == 3 because it's covered by kind == 2

            elif event.kind == 0:
                # starting point
                segA = sweepline.leftof(seg)
                segB = sweepline.rightof(seg)
                if segA is not None and _intersects(segA, seg):
                    return True
                if segB is not None and _intersects(seg, segB):
                    return True

            elif event.kind == 1:
                # ending point
                segA = event.left
                segB = event.right
                if segA is not None and segB is not None and _intersects(segA, segB):
                    return True

    return False

def intersects_sph(CoordString a, CoordString b):
    """ Shamos-Hoey intersection detection algorithm with adaptations for
    spherical coordinates. """

   # corner cases:
   # - simultaneous events          handled
   # - parallel overlapping lines   handled
   # - triple intersections         handled(?)
   # - vertical lines               handled
   # - self-intersections           handled
   #
   # all of the above need tests to verify

    cdef int na = len(a), nb = len(b)
    if a.ring:
        na = na + 1
    if b.ring:
        nb = nb + 1

    # initialize event queue
    #
    # event kind may be:
    # 0 : left point
    # 1 : right point
    # 2 : bottom point on vertical segment
    # 3 : top point on vertical segment
    # 4 : intersection (unused)
    cdef int i = 0
    cdef object event_queue = EventQueuePlaceholder()
    cdef double x0, x1, y0, y1
    cdef tuple seg

    for i in range(na-1):
        x0 = (a.getX(i) + 180.0) % 360.0 - 180.0
        x1 = (a.getX(i+1) + 180.0) % 360.0 - 180.0
        y0 = a.getY(i)
        y1 = a.getY(i+1)
        seg = (x0, y0, x1, y1, 0)
        if x0 == x1:
            if y0 < y1:
                event_queue.insert(x0, Event(seg, 2, 0))
                event_queue.insert(x0, Event(seg, 3, 0))
            else:
                event_queue.insert(x0, Event(seg, 3, 0))
                event_queue.insert(x0, Event(seg, 2, 0))
        elif x0 < x1:
            event_queue.insert(x0, Event(seg, 0, 0))
            event_queue.insert(x1, Event(seg, 1, 0))
        else:
            event_queue.insert(x1, Event(seg, 0, 0))
            event_queue.insert(x0, Event(seg, 1, 0))

    for i in range(nb-1):
        x0 = (b.getX(i) + 180.0) % 360.0 - 180.0
        x1 = (b.getX(i+1) + 180.0) % 360.0 - 180.0
        y0 = b.getY(i)
        y1 = b.getY(i+1)
        seg = (x0, y0, x1, y1, 1)
        if x0 == x1:
            if y0 < y1:
                event_queue.insert(x0, Event(seg, 2, 1))
                event_queue.insert(x0, Event(seg, 3, 1))
            else:
                event_queue.insert(x0, Event(seg, 3, 1))
                event_queue.insert(x0, Event(seg, 2, 1))
        elif x0 < x1:
            event_queue.insert(x0, Event(seg, 0, 1))
            event_queue.insert(x1, Event(seg, 1, 1))
        else:
            event_queue.insert(x1, Event(seg, 0, 1))
            event_queue.insert(x0, Event(seg, 1, 1))

    # begin sweep search
    cdef double x
    cdef double yA
    cdef int event_type, segcolor
    cdef object sweepline = SweepLinePlaceholder(spherical=True)
    cdef Event event = None, ev = None
    cdef tuple segA, segB
    cdef list current_events = [], neighbours = []

    while len(event_queue) != 0:

        # get all simultaneous events
        current_events = []
        flag = 1
        while flag == 1:
            flag, x, event = event_queue.takefirst()
            current_events.append((x, event))

        # for all segments that begin, insert into sweepline
        for x, event in current_events:
            if event.kind == 0:
                sweepline.insert(event.seg, x)

        # for all segments that end, get the neighbours that are not also
        # ending and remove from sweepline
        for _, event in current_events:
            if event.kind == 1:
                segA = event.seg
                segB = event.seg
                while segA is not None and \
                        segA in (ev.seg for x, ev in current_events if ev.kind==1):
                    segA = sweepline.leftof(segA)
                while segB is not None and \
                        segB in (ev.seg for x, ev in current_events if ev.kind==1):
                    segB = sweepline.rightof(segB)
                event.left = segA
                event.right = segB
                sweepline.remove(event.seg)

        # check for intersections due to last set of events
        for x, event in current_events:

            seg = event.seg

            if (event.kind == 2):
                # vertical segment - check whether span crosses any segments
                for segA in sweepline:

                    if seg[4] == segA[4]:
                        continue

                    yA = intersection_meridian(segA[0], segA[2], segA[1], segA[3], x)
                    if (seg[1] <= yA) and (seg[3] >= yA):
                        return True

                    if (seg[3] <= yA) and (seg[1] >= yA):
                        return True

                # can skip kind == 3 because it's covered by kind == 2

            elif event.kind == 0:
                # starting point
                segA = sweepline.leftof(seg)
                segB = sweepline.rightof(seg)
                if segA is not None and _intersects_sph(segA[0], segA[2],
                                                        seg[0], seg[2],
                                                        segA[1], segA[3],
                                                        seg[1], seg[3]):
                    return True
                if segB is not None and _intersects_sph(segB[0], segB[2],
                                                        seg[0], seg[2],
                                                        segB[1], segB[3],
                                                        seg[1], seg[3]):
                    return True

            elif event.kind == 1:
                # ending point
                segA = event.left
                segB = event.right
                if segA is not None and segB is not None \
                        and _intersects_sph(segB[0], segB[2], seg[0], seg[2],
                                            segB[1], segB[3], seg[1], seg[3]):
                    return True

    return False

cdef bool _intersects(tuple seg1, tuple seg2):
    """ Test whether two segments have an intersection. """
    cdef tuple interx
    if seg1[4] == seg2[4]:
        # same geometry
        return False
    if iscollinear(seg1[0], seg1[2], seg2[0], seg2[2],
            seg1[1], seg1[3], seg2[1], seg2[3]):
        return True
    interx = intersection(seg1[0], seg1[2], seg2[0], seg2[2],
                          seg1[1], seg1[3], seg2[1], seg2[3])
    if isnan(interx[0]):
        return False
    return True

#cdef Vector2 _project_gnomonic(Vector2 lonlat, lon0, R):
#    cdef double x = R*tan(lonlat.x - lon0)
#    cdef double y = R*tan(lonlat.y)/cos(lonlat.x-lon0)
#    return Vector2(x, y)
#
#cdef Vector2 _invproject_gnomonic(Vector2 xy, lon0, R):
#    return

cdef bool overlaps(double a0, double a1, double b0, double b1):
    if (a0 <= b0 <= a1) or (a0 <= b1 <= a1) or (b0 <= a0 <= b1) or (b0 <= a1 <= b1):
        return True
    else:
        return False

cdef bool iscollinear(double x0, double x1, double x2, double x3,
                      double y0, double y1, double y2, double y3):
    cdef double rxs = cross2(Vector2(x1-x0, y1-y0), Vector2(x3-x2, y3-y2))
    cdef double rxt = -1.0
    if rxs == 0:
        rxt = cross2(Vector2(x1-x0, y1-y0), Vector2(x3-x0, y3-y0))
        if rxt == 0:
            if (x1-x0) != 0 and overlaps(x0, x1, x2, x3):
                return True
            elif overlaps(y0, y1, y2, y3):
                return True
    return False

cdef bool iscollinear_sph(double x0, double x1, double x2, double x3,
                          double y0, double y1, double y2, double y3):
    cdef Vector3 ep1 = eulerpole(Vector2(x0, y0), Vector2(x1, y1))
    cdef Vector3 ep2 = eulerpole(Vector2(x2, y2), Vector2(x3, y3))
    # TODO: check overlap?
    if ((ep1.x == ep2.x) and (ep1.y == ep2.y) and (ep1.z == ep2.z)) \
            or ((ep1.x == -ep2.x) and (ep1.y == -ep2.y) and (ep1.z == -ep2.z)):
        return True
    else:
        return False

cpdef bool _intersects_sph(double x0, double x1, double x2, double x3,
                           double y0, double y1, double y2, double y3):
    cdef Vector3 ep1 = eulerpole(Vector2(x0, y0), Vector2(x1, y1))
    cdef Vector3 ep2 = eulerpole(Vector2(x2, y2), Vector2(x3, y3))
    cdef Vector2 normal = cart2sph(cross3(ep1, ep2))
    cdef double antipodal_lon = (normal.x + 360) % 360 - 180.0
    if isbetween_inc(x0, normal.x, x1) and isbetween_inc(x2, normal.x, x3):
        return True
    elif isbetween_inc(x0, antipodal_lon, x1) and isbetween_inc(x2, antipodal_lon, x3):
        return True
    else:
        return False

cdef double intersection_meridian(double x0, double x1, double y0, double y1,
                                  double xmeridian):
    """ Return the latitude of intersection of a spherical geodesic across a
    meridian. """
    cdef Vector3 ep1 = eulerpole(Vector2(x0, y0), Vector2(x1, y1))
    cdef Vector3 ep2 = Vector3(sin(M_PI*xmeridian/180.0),
                               cos(M_PI*xmeridian/180.0),
                               0.0)
    cdef Vector2 normal = cart2sph(cross3(ep1, ep2))
    normal.x = (normal.x + 180.0) % 360.0 - 180.0
    normal.y = (normal.y + 90.0) % 180.0 - 90.0
    cdef antipodal_lon = (normal.x + 360) % 360 - 180.0
    if isbetween_inc(x0, normal.x, x1):
        return normal.y
    elif isbetween_inc(x0, antipodal_lon, x1):
        return -normal.y
    else:
        return NAN

cpdef tuple intersection_sph(double x0, double x1, double x2, double x3,
                             double y0, double y1, double y2, double y3):
    """ Compute the spherical intersection between two segments,
    ((x0,y0), (x1,y1)) and ((x2,y2), (x3,y3)).

    Returns (NaN, NaN) if no intersection exists.
    """
    cdef Vector3 ep1 = eulerpole(Vector2(x0, y0), Vector2(x1, y1))
    cdef Vector3 ep2 = eulerpole(Vector2(x2, y2), Vector2(x3, y3))
    cdef Vector2 normal = cart2sph(cross3(ep1, ep2))
    normal.x = (normal.x + 180.0) % 360.0 - 180.0
    normal.y = (normal.y + 90.0) % 180.0 - 90.0
    cdef antipodal_lon = (normal.x + 360) % 360 - 180.0
    cdef Vector2 antipode = Vector2(antipodal_lon, -normal.y)
    if isbetween_inc(x0, normal.x, x1) and isbetween_inc(x2, normal.x, x3):
        return normal.x, normal.y
    elif isbetween_inc(x0, antipode.x, x1) and isbetween_inc(x2, antipode.x, x3):
        return antipode.x, antipode.y
    else:
        return NAN, NAN

cpdef tuple intersection(double x0, double x1, double x2, double x3,
                         double y0, double y1, double y2, double y3):
    """ Compute the planar intersection between two segments,
    ((x0,y0), (x1,y1)) and ((x2,y2), (x3,y3)).

    Returns (NaN, NaN) if no intersection exists.
    """
    cdef double rxs = cross2(Vector2(x1-x0, y1-y0), Vector2(x3-x2, y3-y2))
    if rxs == 0:
        return NAN, NAN

    cdef double t = cross2(Vector2(x2-x0, y2-y0), Vector2(x3-x2, y3-y2)) / rxs
    cdef double u = cross2(Vector2(x2-x0, y2-y0), Vector2(x1-x0, y1-y0)) / rxs
    if (0 < t <= 1) and (0 < u <= 1):
        return x0 + t*(x1-x0), y0 + t*(y1-y0)
    else:
        return NAN, NAN

def count_crossings(double xp, double yp, CoordString coords):
    """ Count the number of times a vertical ray from (xp, yp) crosses a
    line/polygon defined by *coords*
    """
    cdef int n = len(coords)
    cdef int i, cnt = 0
    if not coords.ring:
        n -= 1
    cdef double x0 = coords.getX(0)
    cdef double y0 = coords.getY(0)
    cdef double x1, y1
    for i in range(1, n):
        x1 = coords.getX(i)
        y1 = coords.getY(i)
        if intersects_cn(xp, yp, x0, x1, y0, y1) == 1:
            cnt += 1
        x0 = x1
        y0 = y1
    return cnt

cdef int intersects_cn(double xp, double yp, double x0, double x1, double y0, double y1):
    """ Test whether a vertical ray emanating up from a point (xp, yp) crosses
    a line segment [(x0, y0), (x1, y1)]. Used to implement a crossing number
    membership test.
    """
    cdef double m, y

    if x0 != x1:
        m = (y1-y0) / (x1-x0)
    else:
        return 0

    y = y0 + m * (xp - x0)

    if y < yp:
        return 0

    cdef int iswithinx = 0
    cdef int iswithiny = 0

    if m > 0.0 and isbetween_incr(y0, y, y1):
        iswithiny = 1
    elif isbetween_incl(y0, y, y1):
        iswithiny = 1
    elif abs(y0-y1) < 1e-15 and abs(y-y0) < 1e-15:
        iswithiny = 1

    if isbetween_incr(x0, xp, x1):
        iswithinx = 1

    return iswithinx*iswithiny




import numpy as np
cimport numpy as np
cimport cython

DTYPE_float64 = np.float64
ctypedef np.float64_t DTYPE_float64_t

DTYPE_int32 = np.int32
ctypedef np.int32_t DTYPE_int32_t

@cython.cdivision(True)
def get_positions_vec(tuple T, double[:] x not None, double[:] y not None):
    """ Return the row and column indices according to grid transform T for
    points in vectors x and y. """
    cdef int i = 0
    cdef int n = len(x)
    cdef double x0, y0, dx, dy, sx, sy
    cdef double i_, j_
    cdef np.ndarray[np.float64_t] I = np.empty(n)
    cdef np.ndarray[np.float64_t] J = np.empty(n)
    x0 = T[0]
    y0 = T[1]
    dx = T[2]
    dy = T[3]
    sx = T[4]
    sy = T[5]
    while i != n:
        j_ = (dy*x[i] - dy*x0 + sx*y0 - sx*y[i]) / (dx*dy - sx*sy)
        i_ = (y[i] - y0 - j_*sy) / dy
        J[i] = j_ - 0.5
        I[i] = i_ - 0.5
        i += 1
    return I, J

def sample_bilinear_uint(double[:] I not None, double[:] J not None,
                         np.uint16_t[:,:] Z not None, unsigned int na):
    cdef int cnt = 0
    cdef int L = len(I)
    cdef double i = 0.0, j = 0.0
    cdef int i0 = 0, i1 = 0, j0 = 0, j1 = 0
    cdef int m, n
    m = Z.shape[0]
    n = Z.shape[1]
    cdef np.ndarray[np.uint16_t] out = np.empty(L, dtype=np.uint16)

    for cnt in range(L):
        i = I[cnt]
        j = J[cnt]
        if (i % 1 != 0):
            i0 = int(i // 1.0)
            i1 = int(i0 + 1)
        elif (i != 0):
            i0 = int(i - 1.0)
            i1 = int(i)
        else:
            i0 = int(i)
            i1 = int(i + 1.0)

        if (j % 1 != 0):
            j0 = int(j // 1.0)
            j1 = int(j0 + 1)
        elif (j != 0):
            j0 = int(j - 1.0)
            j1 = int(j)
        else:
            j0 = int(j)
            j1 = int(j + 1.0)

        if i0 >= 0 and i1 < m and j0 >= 0 and j1 < n:
            out[cnt] = int(float(Z[i0,j0]) * (i1-i) * (j1-j)
                         + float(Z[i1,j0]) * (i-i0) * (j1-j) \
                         + float(Z[i0,j1]) * (i1-i) * (j-j0) \
                         + float(Z[i1,j1]) * (i-i0) * (j-j0))
        else:
            out[cnt] = na
    return out

def sample_bilinear_int(double[:] I not None, double[:] J not None,
                         np.int32_t[:,:] Z not None, int na):
    cdef int cnt = 0
    cdef int L = len(I)
    cdef double i = 0.0, j = 0.0
    cdef int i0 = 0, i1 = 0, j0 = 0, j1 = 0
    cdef int m, n
    m = Z.shape[0]
    n = Z.shape[1]
    cdef np.ndarray[np.int32_t] out = np.empty(L, dtype=np.int32)

    for cnt in range(L):
        i = I[cnt]
        j = J[cnt]
        if (i % 1 != 0):
            i0 = int(i // 1.0)
            i1 = int(i0 + 1)
        elif (i != 0):
            i0 = int(i - 1.0)
            i1 = int(i)
        else:
            i0 = int(i)
            i1 = int(i + 1.0)

        if (j % 1 != 0):
            j0 = int(j // 1.0)
            j1 = int(j0 + 1)
        elif (j != 0):
            j0 = int(j - 1.0)
            j1 = int(j)
        else:
            j0 = int(j)
            j1 = int(j + 1.0)

        out[cnt] = int(float(Z[i0,j0]) * (i1-i) * (j1-j)
                     + float(Z[i1,j0]) * (i-i0) * (j1-j) \
                     + float(Z[i0,j1]) * (i1-i) * (j-j0) \
                     + float(Z[i1,j1]) * (i-i0) * (j-j0))
    return out

def sample_bilinear_double(double[:] I not None, double[:] J not None,
                           np.float64_t[:,:] Z not None, double na):
    cdef int cnt = 0
    cdef int L = len(I)
    cdef double i = 0.0, j = 0.0
    cdef int i0 = 0, i1 = 0, j0 = 0, j1 = 0
    cdef int m, n
    m = Z.shape[0]
    n = Z.shape[1]
    cdef np.ndarray[np.float64_t] out = np.empty(L)

    for cnt in range(L):
        i = I[cnt]
        j = J[cnt]
        if (i % 1 != 0):
            i0 = int(i // 1.0)
            i1 = int(i0 + 1)
        elif (i != 0):
            i0 = int(i - 1.0)
            i1 = int(i)
        else:
            i0 = int(i)
            i1 = int(i + 1.0)

        if (j % 1 != 0):
            j0 = int(j // 1.0)
            j1 = int(j0 + 1)
        elif (j != 0):
            j0 = int(j - 1.0)
            j1 = int(j)
        else:
            j0 = int(j)
            j1 = int(j + 1.0)

        if i0 >= 0 and i1 < m and j0 >= 0 and j1 < n:
            out[cnt] = Z[i0,j0]*(i1-i)*(j1-j) + Z[i1,j0]*(i-i0)*(j1-j) \
                     + Z[i0,j1]*(i1-i)*(j-j0) + Z[i1,j1]*(i-i0)*(j-j0)
        else:
            out[cnt] = na
    return out

@cython.cdivision(True)
@cython.wraparound(False)
def fillarray_double(double[:,:] array not None,
                     int[:] I not None,
                     int[:] J not None,
                     double[:] Z not None,
                     double nodata_value):
    """ fillarray_double places values from *Z* into 2d *array* at the indices
    *I*, *J*. *array* should be a zero array.

    Returns 0 on success and 1 on failure.
    """
    cdef int ny, nx
    cdef int i, j, idx
    cdef double z
    cdef np.ndarray[DTYPE_int32_t, ndim=2] counts

    if (len(I) != len(J)) or (len(I) != len(Z)):
        return 1

    ny = array.shape[0]
    nx = array.shape[1]
    counts = np.zeros([ny, nx], dtype=DTYPE_int32)

    for idx in range(len(I)):
        i = I[idx]
        j = J[idx]
        array[i,j] += Z[idx]
        counts[i,j] += 1

    for i in range(ny):
        for j in range(nx):
            if counts[i,j] != 0:
                array[i,j] = array[i,j] / counts[i,j]
            else:
                array[i,j] = nodata_value
    return 0




"""
karta.vector

Classes and functions for vector data
"""

import numpy
import numbers
numbers.Integral.register(numpy.integer)

from . import geometry
from . import quadtree
from . import rtree
from . import table
from . import _geojson as geojson
from . import _gpx as gpx
from . import _shp as shp
#from . import redblack

from .geometry import (Geometry, Point, Line, Polygon,
                       Multipoint, Multiline, Multipolygon,
                       multipart_from_singleparts, merge_multiparts)
from .read import (from_shape, read_geojson, read_shapefile,
                   read_gpx_waypts, read_gpx_tracks)
from .table import Table
from ._shp import write_shapefile

__all__ = ["geometry", "geojson", "gpx",
           "Point", "Line", "Polygon",
           "Multipoint", "Multiline", "Multipolygon",
           "multipart_from_singleparts",
           "Table",
           "read_geojson", "read_shapefile", "write_shapefile"]





def cache_decorator(key):
    """ Returns a method decorator that stores the result of a function
    invocation under `self._cache[key]`

    Does not update cache when *crs* is specified. """
    def wrapping_func(f):
        def replacement_func(self, *args, **kwargs):
            if (key in self._cache and
                    (args == self._cache[key]["args"]) and
                    (kwargs == self._cache[key]["kwargs"])):
                ret = self._cache[key]["output"]
            else:
                ret = f(self, *args, **kwargs)
                self._cache[key] = dict(output=ret, args=args, kwargs=kwargs)
            return ret
        return replacement_func
    return wrapping_func




""" Cython wrapper for rtree """

cdef extern from "rtree.h":

    cdef enum Strategy:
        LINEAR, QUADRATIC

    cdef enum NodeType:
        LEAF, NONLEAF

    cdef struct BoundingBox:
        float xmin
        float ymin
        float xmax
        float ymax
    ctypedef BoundingBox Bbox

    cdef struct RTreeNode:
        int count
        Bbox* bbox
    ctypedef RTreeNode Node

    cdef struct PointerPool:
        int size
        int count
    ctypedef PointerPool Pool

    Node* rt_new_node(NodeType, Strategy, int, Node*)
    Node* rt_insert(Node*, Bbox*, int)
    Bbox* rt_new_bbox()
    void rt_free(Node*)
    void print_bbox(Bbox*)
    Pool *rt_search_within(Node*, Bbox*, int)
    Pool *rt_search_overlapping(Node*, Bbox*, int)

    char *pool_pop(Pool*, int)
    void pool_destroy(Pool*)

cdef class RTree:
    cdef int count
    cdef Node* root

    def __init__(self, list geometries, maxchildren=50):
        self.root = NULL
        cdef int i = 0
        cdef object geom
        cdef Node* root

        root = rt_new_node(LEAF, LINEAR, maxchildren, NULL)
        for geom in geometries:
            if not hasattr(geom, "bbox"):
                raise AttributeError("cannot construct R-tree index from items "
                                     "missing a `bbox` attribute")
            _bb = geom.bbox
            bb = rt_new_bbox()
            bb.xmin = _bb[0]
            bb.ymin = _bb[1]
            bb.xmax = _bb[2]
            bb.ymax = _bb[3]
            root = rt_insert(root, bb, i)
            i += 1

        self.count = i
        self.root = root
        return

    def __dealloc__(self):
        if self.root != NULL:
            rt_free(self.root)

    @property
    def bbox(self):
        return (self.root.bbox.xmin, self.root.bbox.ymin,
                self.root.bbox.xmax, self.root.bbox.ymax)

    def search_within(self, bbox, int max_results=-1):
        """ Return a list of geometries that are within a bounding box. """
        cdef Pool* result
        cdef Bbox* bb = rt_new_bbox()
        cdef list out = []

        bb.xmin = bbox[0]
        bb.ymin = bbox[1]
        bb.xmax = bbox[2]
        bb.ymax = bbox[3]

        result = rt_search_within(self.root, bb, max_results)
        while result.count != 0:
            out.append((<int*> pool_pop(result, result.count-1))[0])
        pool_destroy(result)
        return out

    def search_overlapping(self, bbox, int max_results=-1):
        """ Return a list of geometries that are within a bounding box. """
        cdef Pool* result
        cdef Bbox* bb = rt_new_bbox()
        cdef list out = []

        bb.xmin = bbox[0]
        bb.ymin = bbox[1]
        bb.xmax = bbox[2]
        bb.ymax = bbox[3]

        result = rt_search_overlapping(self.root, bb, max_results)
        while result.count != 0:
            out.append((<int*> pool_pop(result, result.count-1))[0])
        pool_destroy(result)
        return out




""" IO interface to GeoTiffs using GDAL. """

import struct
import sys
from math import ceil
import numpy as np
from .band import CompressedBand
from .. import errors

import osgeo.gdal
import osgeo.osr
import osgeo.gdalconst as gc
osgeo.gdal.UseExceptions()

ALL = -1

class GdalFileBand(object):
    """ Read-only raster Band interface the reads data from a disk-bound
    datasource.
    """
    def __init__(self, band, dataset):
        """
        Parameters
        ----------
        band : osgeo.gdal.Band
        dataset : osgeo.gdal.Dataset
        """
        self.gdalband = band
        self.dataset = dataset
        return

    def __del__(self):
        self.dataset = None
        self.gdalband = None

    def getblock(self, yoff, xoff, ny, nx):
        # Note that GDAL uses the alternative x,y convention
        grid_ny, grid_nx = self.size
        chunk = self.gdalband.ReadAsArray(xoff, grid_ny - yoff - ny, nx, ny)
        if chunk is None:
            raise IOError("failure reading slice from GDAL backend")
        return chunk[::-1]

    def setblock(self, yoff, xoff, array):
        raise NotImplementedError()

    def __iter__(self):
        for i in range(self.dataset.RasterYSize):
            yield self[i,:]

    @property
    def size(self):
        return (self.dataset.RasterYSize, self.dataset.RasterXSize)

    @property
    def dtype(self):
        return np.dtype(numpy_dtype(self.gdalband.DataType))

def SRS_from_WKT(s):
    """ Return Proj.4 string, semimajor axis, and flattening """
    sr = osgeo.osr.SpatialReference()
    try:
        sr.ImportFromWkt(s)
    except RuntimeError:
        sr = None
    return sr

NUMPY_DTYPE = {"Byte": np.uint8,
               "UInt16": np.uint16,
               "Int16": np.int16,
               "UInt32": np.uint32,
               "Int32": np.int32,
               "Float32": np.float32,
               "Float64": np.float64,
               "CInt16": np.complex64,
               "CInt32": np.complex64,
               "CFloat32": np.complex64,
               "CFloat64": np.complex64}

def numpy_dtype(dt_int):
    """ Return a numpy dtype that matches the band data type. """
    name = osgeo.gdal.GetDataTypeName(dt_int)
    if name in NUMPY_DTYPE:
        return NUMPY_DTYPE[name]
    else:
        raise TypeError("GDAL data type {0} unknown to karta".format(dt_int))

def gdal_type(dtype):
    """ Return a GDAL type that most closely matches numpy dtype

    Notes
    -----
    Returns GDT_Int32 for np.int64, which may result in overflow.
    """
    if dtype == np.uint8:
        return osgeo.gdal.GDT_Byte
    elif dtype == np.uint16:
        return osgeo.gdal.GDT_UInt16
    elif dtype == np.int8:
        return osgeo.gdal.GDT_Byte      # transform -127 -- 127 to 0 -- 255
    elif dtype == np.int16:
        return osgeo.gdal.GDT_Int16
    elif (dtype == np.int32) or (dtype == np.int64):
        return osgeo.gdal.GDT_Int32
    elif dtype == np.float32:
        return osgeo.gdal.GDT_Float32
    elif dtype == np.float64:
        return osgeo.gdal.GDT_Float64
    elif dtype == np.complex64:
        return osgeo.gdal.GDT_CFloat64
    else:
        raise TypeError("GDAL equivalent to type {0} unknown".format(dtype))

def read(fnm, in_memory, ibands=ALL, bandclass=CompressedBand):
    """ Read a GeoTiff file and return a numpy array and a dictionary of header
    information.

    Parameters
    ----------
    fnm : str
        input datasource
    in_memory : boolean
        indicates whether array should be read fully into memory
    ibands : int or list of ints
        band number (1...)
    bandclass : karta.raster.band class
        if *in_memory* is `False`, use this class for band storage

    Returns an band object and a dictionary of metadata
    """
    hdr = dict()
    dataset = osgeo.gdal.Open(fnm, gc.GA_ReadOnly)

    if ibands == ALL:
        ibands = list(range(1, dataset.RasterCount+1))
    elif not hasattr(ibands, "__iter__"):
        ibands = [ibands]

    try:
        hdr["nx"] = dataset.RasterXSize
        hdr["ny"] = dataset.RasterYSize

        transform = dataset.GetGeoTransform()
        if transform is not None:
            hdr["dx"] = transform[1]
            hdr["dy"] = transform[5]
            hdr["xulcorner"] = transform[0]
            hdr["yulcorner"] = transform[3]
            hdr["sx"] = transform[2]
            hdr["sy"] = transform[4]
        else:
            raise AttributeError("No GeoTransform in geotiff file")

        sr = SRS_from_WKT(dataset.GetProjectionRef())
        if sr is not None:
            hdr["srs"] = {"proj4": sr.ExportToProj4(),
                          "semimajor": sr.GetSemiMajor(),
                          "flattening": sr.GetInvFlattening(),
                          "name": sr.GetAttrValue('PROJCS')}
        else:
            hdr["srs"] = {"proj4": "",
                          "semimajor": 6370997.0,
                          "flattening": 1.0 / 298.257223563,
                          "name": "NA"}

        max_dtype = 0
        rasterbands = [dataset.GetRasterBand(i) for i in ibands]
        hdr["nodata"] = rasterbands[0].GetNoDataValue()
        nx = rasterbands[0].XSize
        ny = rasterbands[0].YSize
        if rasterbands[0].DataType > max_dtype:
            max_dtype = rasterbands[0].DataType

        if in_memory:
            dtype = numpy_dtype(rasterbands[0].DataType)
            bands = [bandclass((ny, nx), dtype) for _ in ibands]
            for i, rb in enumerate(rasterbands):
                _arr = rb.ReadAsArray(buf_obj=np.empty([ny, nx], dtype=dtype))
                if _arr is None:
                    raise IOError("error reading GDAL band {}".format(i+1))
                bands[i].setblock(0, 0, _arr.squeeze()[::-1])
        else:
            bands = [GdalFileBand(rb, dataset) for rb in rasterbands]

    finally:
        if in_memory:
            dataset = None
    return bands, hdr

def srs_from_crs(crs):
    srs = osgeo.osr.SpatialReference()
    # SpatialReference can't parse 'lonlat'
    proj4 = crs.get_proj4().replace("lonlat", "latlong")
    srs.ImportFromProj4(proj4)
    return srs

def write(fnm, grid, compress=None, tiled=False, **kw):
    """ Write a grid-like object to *fnm* """
    co = []
    if compress == "LZW":
        co.append("COMPRESS=LZW")
    elif compress == "PACKBITS":
        co.append("COMPRESS=PACKBITS")
    elif compress == "DEFLATE":
        co.append("COMPRESS=DEFLATE")
    elif compress == "LZMA":
        co.append("COMPRESS=LZMA")

    if tiled:
        co.append("TILED=YES")

    for k, v in kw.items():
        co.append("{0}={1}".format(k,v))

    driver = osgeo.gdal.GetDriverByName("GTiff")
    ny, nx = grid.size
    dataset = driver.Create(fnm, nx, ny, len(grid.bands), gdal_type(grid.values.dtype), co)
    t = grid.transform
    dataset.SetGeoTransform([t[0] + ny*t[4], t[2], -t[4],
                             t[1] + ny*t[3], t[5], -t[3]])
    srs = srs_from_crs(grid.crs)
    dataset.SetProjection(srs.ExportToWkt())
    for i, _ in enumerate(grid.bands):
        band = dataset.GetRasterBand(i+1)
        band.SetNoDataValue(grid.nodata)

        tmp = grid[:,:,i]
        band.WriteArray(tmp)
        grid[:,:,i] = tmp[::-1]
    band = None
    dataset = None
    return




""" Cython wrapper quadtree """

# from .coordstring import Coordstring
# from .coordstring cimport Coordstring

cdef extern from "quadtree.h":

    ctypedef enum NodeType:
        LEAF, NONLEAF

    ctypedef enum Quadrant:
        UPPERLEFT, UPPERRIGHT, LWERLEFT, LOWERRIGHT

    struct BoundingBox:
        double xmin
        double ymin
        double xmax
        double ymax
    ctypedef BoundingBox Bbox

    struct PositionStruct:
        int id
        double x
        double y
    ctypedef PositionStruct Position

    struct QuadTreeLeafNode:
        Bbox *bbox
    ctypedef QuadTreeLeafNode LeafNode

    struct QuadTreeNonleafNode:
        Bbox *bbox
    ctypedef QuadTreeNonleafNode NonleafNode

    union NodePtrUnionTag:
        QuadTreeLeafNode *leafnode
        QuadTreeNonleafNode *nonleafnode
    ctypedef NodePtrUnionTag NodePtrUnion

    cdef struct PointerPool:
        int size
        int count
    ctypedef PointerPool Pool

    Position *qt_new_position(int, double, double)
    void qt_free_position(Position*)

    Bbox *qt_new_bbox(double, double, double, double)
    void qt_free_bbox(Bbox*)

    LeafNode *qt_new_leaf(int, Bbox*)
    NonleafNode *qt_new_nonleaf(Bbox*)
    NonleafNode *qt_insert(NodePtrUnion, Position, int*)
    void qt_free_node(NodePtrUnion)
    Pool *qt_search_within(NodePtrUnion, Bbox*)
    char *pool_pop(Pool*, int)

cdef class QuadTree:
    cdef int count
    cdef NodePtrUnion root
    cdef readonly dict duplicates

    def __cinit__(self, points, int leaf_capacity=50):
        cdef NonleafNode *retnode
        cdef Position *pos
        cdef Bbox *bbox
        cdef double x, y
        cdef double xmin, ymin, xmax, ymax
        cdef int i = 0
        cdef int idup = 0
        cdef dict duplicates = {}

        xmin, ymin, xmax, ymax = points.bbox
        bbox = qt_new_bbox(xmin, ymin, xmax, ymax)
        self.root.leafnode = qt_new_leaf(leaf_capacity, bbox)

        while i != len(points):
            x, y = points[i][:2]
            pos = qt_new_position(i, x, y)
            retnode = qt_insert(self.root, pos[0], &idup)
            if idup != -1:
                if idup in duplicates:
                    duplicates[idup].append(i)
                else:
                    duplicates[idup] = [i]
            elif (retnode != NULL):
                self.root.nonleafnode = retnode
            qt_free_position(pos)
            i += 1

        self.count = i
        self.duplicates = duplicates
        return

    def __dealloc__(self):
        if not (self.root.leafnode == NULL):
            qt_free_node(self.root)

    def __len__(self):
        return self.count

    @property
    def bbox(self):
        cdef NonleafNode *node = self.root.nonleafnode
        cdef Bbox *bbox = node.bbox
        cdef double xmin = bbox.xmin
        cdef double ymin = bbox.ymin
        cdef double xmax = bbox.xmax
        cdef double ymax = bbox.ymax
        return (xmin, ymin, xmax, ymax)

    def search_within(self, double xmin, double ymin, double xmax, double ymax):
        cdef Bbox *bbox = qt_new_bbox(xmin, ymin, xmax, ymax)
        cdef Pool *results = qt_search_within(self.root, bbox)
        cdef int *idx
        cdef list out = []

        while results.count != 0:
            idx = <int*> pool_pop(results, results.count-1)
            if idx[0] in self.duplicates:
                out.extend(self.duplicates[idx[0]])
            out.append(idx[0])
        qt_free_bbox(bbox)
        return out




import picogeojson
from .utilities import _as_nested_lists, _reproject_nested
from ..crs import LonLatWGS84

def crs_from_urn(urn):
    return {'type': 'name', 'properties': {'name': urn}}

def crs_from_karta(crs):
    if hasattr(crs, "jsonhref") and hasattr(crs, "jsontype"):
        return {'type': 'link', 'properties': {'href': crs.jsonname, 'type': crs.jsontype}}
    elif hasattr(crs, "jsonname"):
        return {'type': 'name', 'properties': {'name': crs.jsonname}}
    else:
        # NOTE: make sure CRS gets written to file by caller
        return {'type': 'link', 'properties': {'href': "", 'type': 'proj4'}}

class GeoJSONOutMixin(object):
    """ Mixin class to be added to geometry objects, adding geojson
    functionality.
    """

    _serializer = picogeojson.Serializer(antimeridian_cutting=False,
                                         enforce_poly_winding=False)

    @staticmethod
    def _as_named_tuple(geom, **kwargs):
        """ Convert one or more Geometry instances to GeoJSON-structured named tuples.
        Parameters
        ----------
        *geoms : subtypes of karta.Geometry
            karta vector geometries to convert
        urn : str, optional
            URN defining a specific CRS to use
        Returns
        -------
        Either a Feature or a FeatureCollection
        Raises
        ------
        TypeError
            if one or more of geoms has an unrecognized `_geotype` attribute
        """
        if kwargs.get("urn", None) is not None:
            crs = crs_from_urn(kwargs["urn"])
        else:
            crs = crs_from_karta(geoms[0].crs)

        if geom._geotype == "Point":
            g = picogeojson.Point(geom.vertex, crs)
        elif geom._geotype == "Line":
            g = picogeojson.LineString(_as_nested_lists(geom.vertices), crs)
        elif geom._geotype == "Polygon":
            verts = [_as_nested_lists(geom.vertices_ring)]
            for sub in geom.subs:
                verts.append(_as_nested_lists(sub.vertices))
            g = picogeojson.Polygon(verts, crs)
        elif geom._geotype == "Multipoint":
            g = picogeojson.MultiPoint(_as_nested_lists(geom.vertices), crs)
        elif geom._geotype == "Multiline":
            g = picogeojson.MultiLineString(_as_nested_lists(geom.vertices), crs)
        elif geom._geotype == "Multipolygon":
            g = picogeojson.MultiPolygon(_as_nested_lists(geom.vertices_ring), crs)
        else:
            raise TypeError("unhandled type: {0}".format(type(geom)))

        data = {}
        if hasattr(geom, "data"):
            for field in geom.data.fields:
                data[field] = geom.d[field]

        properties = geom.properties
        properties.update(data)
        return picogeojson.Feature(g, properties)

    def as_geojson(self, indent=None, urn=None, force_wgs84=True):
        """ Output representation of internal data as a GeoJSON string.

        Parameters
        ----------
        indent : int, optional
            indentation of generated GeoJSON (default 2)
        force_wgs84 : bool, optional
            Forces output to use geographical coordinates with the WGS84 datum,
            as recommended by the GeoJSON draft specification
            (https://datatracker.ietf.org/doc/draft-ietf-geojson/).
            If *urn* is not set, "urn:ogc:def:crs:OGC:1.3:CRS84" is used.
            (default True)
        urn : str, optional
            overrides GeoJSON CRS with provided URN string
        """
        if force_wgs84 and (urn is None):
            urn = "urn:ogc:def:crs:OGC:1.3:CRS84"
        if force_wgs84 and (self.crs != LonLatWGS84):
            kw = dict(properties=self.properties, crs=LonLatWGS84)
            if hasattr(self, "data"):
                kw["data"] = self.data
            if hasattr(self, "vertices"):
                vertices = _reproject_nested(self.vertices, self.crs, LonLatWGS84)
            else:
                vertices = _reproject(self.vertex, self.crs, LonLatWGS84)
            geo = type(self)(vertices, **kw)
        else:
            geo = self

        return self._serializer(self._as_named_tuple(geo, urn=urn),
                                indent=indent)

    def to_geojson(self, f, indent=None, urn=None, force_wgs84=True):
        """ Write data as a GeoJSON string to a file-like object `f`.

        Parameters
        ----------
        f : str or file-like object
            file to receive GeoJSON string
        indent : int, optional
            indentation of generated GeoJSON (default None)
        force_wgs84 : bool, optional
            Forces output to use geographical coordinates with the WGS84 datum,
            as recommended by the GeoJSON draft specification
            (https://datatracker.ietf.org/doc/draft-ietf-geojson/).
            If *urn* is not set, "urn:ogc:def:crs:OGC:1.3:CRS84" is used.
            (default True)
        urn : str, optional
            overrides GeoJSON CRS with provided URN string
        """
        try:
            if not hasattr(f, "write"):
                fobj = open(f, "w")
            else:
                fobj = f
            fobj.write(self.as_geojson(indent=indent, urn=urn,
                                       force_wgs84=force_wgs84))
        finally:
            if not hasattr(f, "write"):
                fobj.close()
        return




import sys
from ez_setup import use_setuptools
use_setuptools()
from os.path import exists
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext

sys.path.append("karta")
from version import __version__ as VERSION

class build_ext(_build_ext):

    # solution taken from http://stackoverflow.com/questions/19919905/ \
    #   how-to-bootstrap-numpy-installation-in-setup-py
    def finalize_options(self):
        """ Unsets __NUMPY_SETUP__ so that the get_include function can be used
        """
        _build_ext.finalize_options(self)

        # Hack: unset __NUMPY_SETUP__ so that it can be imported
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

        # Add include_dirs to each extension so that Cython will find them
        for extension in self.extensions:
            extension.include_dirs = self.include_dirs

        # Attempt to compile with Cython. Fallback on local C sources.
        #
        # Check that sources exist for each extension, either from Cython for
        # from having the C-code sitting around.
        #
        # Add a _needs_stub attribute to each extension for setuptools.
        try:
            from Cython.Build import cythonize
            _needs_stub = [ext._needs_stub for ext in self.extensions]
            self.extensions = cythonize(self.extensions)
            for ns, ext in zip(_needs_stub, self.extensions):
                ext._needs_stub = ns

        except ImportError:
            print("Cython not imported")
            print("If C sources exist in the build directory, they will be used")

            for ext in self.extensions:
                ext.sources = [f.replace(".pyx", ".c") for f in ext.sources]

        for ext in self.extensions:
            for src in ext.sources:
                if not exists(src):
                    raise Exception("Extension source code not found ({0})\n"
                                    "Cython required for install".format(src))
        return

# File extension is added to sources at overloaded build_ext.run()
extensions = [
        Extension("karta.raster.crfuncs", ["karta/raster/crfuncs.pyx"]),

        Extension("karta.vector.coordstring", ["karta/vector/coordstring.pyx"]),

        Extension("karta.vector.vectorgeo", ["karta/vector/vectorgeo.pyx"],
                  extra_compile_args=["-std=c99"]),

        Extension("karta.vector.dateline", ["karta/vector/dateline.pyx"]),

        Extension("karta.vector.intersection", ["karta/vector/intersection.pyx"]),

        Extension("karta.vector.convexhull", ["karta/vector/convexhull.pyx"]),

        Extension("karta.vector.contains", ["karta/vector/contains.pyx"]),

        Extension("karta.vector.quadtree", ["karta/vector/quadtree.pyx"],
                        extra_compile_args=["-std=c99"]),

        Extension("karta.vector.rtree", ["karta/vector/rtree.pyx"],
                  extra_compile_args=["-std=c99"])
        ]

setup(
    name = "karta",
    version = VERSION,
    setup_requires = ["numpy>=1.10"],
    install_requires = ["numpy>=1.10",
                        "pyproj>=1.9",
                        "gdal>=1.10",
                        "blosc>=1.2.8",
                        "picogeojson==0.4.1"],
    author = "Nat Wilson",
    author_email = "natw@fortyninemaps.com",
    packages = ["karta", "karta.vector", "karta.raster"],
    url = "http://www.fortyninemaps.com/karta.html",
    description = "Geospatial analysis in Python",
    long_description = """
|Build Status| |Build status| |Coverage Status|

.. figure:: https://raw.githubusercontent.com/fortyninemaps/karta/gh-pages/images/karta_logo.png
   :alt: Karta

   Karta

*Karta* is a package for spatial analysis in Python. It simplifies
geospatial data processing by providing efficient generic classes for
vector and raster data sources, as well as a selection of analysis
functions.

Vector data types
-----------------

Data are represented as ``Point``, ``Line``, ``Polygon``,
``Multipoint``, ``Multiline``, and ``Multipolygon`` instances.

All data contain a ``.crs`` member encoding coordinate reference
information. All vector geometries possess a ``.properties`` dict
containing free-form metadata. *Multipart* geometries additionally
possess a ``.data`` member which is a simple typed table-like data
structure.

Geometries implement methods for computing distances, directions, and
spatial characteristics. *Multipart* geometries support fast spatial
indexing through quadtrees and r-trees.

GeoJSON and ESRI shapefile formats are supported for reading and
writing. Experimental support for GPX XML files is in the
``karta.vector.gpx`` submodule.

Vector geometries implement the Python ```__geo_interface__``
attribute <https://gist.github.com/sgillies/2217756>`__ for vector
geometries. This permits data to be exchanged between *Karta* and
external modules that also implement ``__geo_interface__`` (e.g.
`shapely <https://github.com/Toblerity/Shapely>`__,
`fastkml <https://fastkml.readthedocs.org/en/latest/>`__).

Raster data types
-----------------

The primary raster data type is the ``RegularGrid``, which represents
one or more 2-d arrays of pixels spaced via an affine transformation.
``RegularGrids`` are backed by one of several ``Band`` implementations,
with the default implementation using the
`blosc <http://www.blosc.org/>`__ compression library for efficient
in-memory storage. There is experimental support for disk-backed storage
via GDAL.

Grids may be queried, resampled, sliced, masked, and merged. Arbitrary
array-based functions may be mapped to raster data with
``RegularGrid.apply()``. Raster functions including slope, gradient, and
hillshade are in the ``karta.raster.misc`` submodule.

GeoTIFF images are the primary supported format, however ESRI ASCII
grids may also be used (with limitations due to the format).

Coordinate reference systems
----------------------------

Data in Karta is referenced to positions on earth via ``CRS`` objects
that implement projection and geodetic methods. Coordinate reference
systems may be either geographical or projected.

**Geographical CRS** objects return spatial relationships in terms of
the true computed distances and azimuths on a spherical or ellipsoidal
Earth.

**Projected CRS** objects (e.g. UTM, Polar Stereographic, and Web
Mercator) return spatial relationships in terms of a flat plane,
dependent on the projection.

Examples
--------

Read or create vector geometries:

.. code:: python

    point = Point((-130.0, 52.0), crs=LonLatWGS84)
    line = read_geojson("linedata.json")
    polygon = Polygon([(-515005.78, -1301130.53),
                       (-579174.89, -1282271.94),
                       (-542977.83, -1221147.82),
                       (-437864.05, -1251641.55),
                       (-438160.72, -1252421.48),
                       (-437961.28, -1285314.00)],
                       crs=NSIDCNorth)

Perform simple queries:

.. code:: python

    point2 = Point((-25.0, 48.0), crs=LonLatWGS84)
    point.distance(point2)          # Distance in geographical units
    line.intersects(polygon)        # True or False
    ch = polygon.convex_hull()      # Returns a new polygon
    ch.to_shapefile("poly.shp")

Load and manipulate raster data:

.. code:: python

    grid = read_gtiff("landsat_scene.tif")  # Leverages GDAL
    grid.profile(line)              # Collect data along a line
    grid.resample(500.0, 500.0)     # Return a grid resampled at a new resolution

Installation
------------

*Karta* currently supports Python 2.7 and Python 3.4+.

The easiest way to install is via ``pip``. Installation requires a recent
version of ``setuptools``.

::

    pip install -U setuptools
    pip install karta

Building from source
~~~~~~~~~~~~~~~~~~~~

Building from source requires Cython and a C99-compliant compiler:

::

    pip install Cython

Then, clone the repository and install:

::

    git clone https://github.com/fortyninemaps/karta.git karta
    cd karta/
    python setup.py build


Documentation
-------------

See the `online
manual <http://www.fortyninemaps.com/kartadocs/introduction.html>`__,
the
`tutorial <http://www.fortyninemaps.com/kartadocs/_static/tutorial.html>`__,
or read the `API
documentation <http://www.fortyninemaps.com/kartadocs/reference.html>`__.

Contributing
------------

Bug reports, feature requests, and pull requests are welcome.

Run unit tests with ``python tests/runtests.py``.

The manual is built using `Sphinx <http://sphinx-doc.org/>`__ and
requires `numpydoc <https://github.com/numpy/numpydoc>`__.

.. |Build Status| image:: https://travis-ci.org/fortyninemaps/karta.svg?branch=master
   :target: https://travis-ci.org/fortyninemaps/karta
.. |Build status| image:: https://ci.appveyor.com/api/projects/status/viiimwp5pu7ff2bp?svg=true
   :target: https://ci.appveyor.com/project/njwilson23/karta
.. |Coverage Status| image:: https://coveralls.io/repos/github/fortyninemaps/karta/badge.svg?branch=master
   :target: https://coveralls.io/github/fortyninemaps/karta?branch=master
""",
    classifiers = ["Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.4",
                   "Programming Language :: Python :: 3.5",
                   "Programming Language :: Python :: 3.6",
                   "Topic :: Scientific/Engineering",
                   "Topic :: Scientific/Engineering :: GIS",
                   "License :: OSI Approved :: MIT License"],
    license = "MIT License",
    ext_modules = extensions,
    cmdclass = {"build_ext": build_ext},
)




class GeometryError(Exception):
    """ Indicates invalid geometrical operations """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message

class CRSError(Exception):
    """ Indicates invalid CRS parameters or operations """
    def __init__(self, message=''):
        self.message = message

class NoIntersection(Exception):
    """ Indicates that no intersection exists between segments """
    def __init__(self, message=''):
        self.message = message

class GridError(Exception):
    """ Indicates invalid Grid initialization, referencing, or transformation """
    def __init__(self, message=''):
        self.message = message
    def __str__(self):
        return self.message




""" Functions for reading raster data sources as RegularGrid objects """
import numpy as np
from .grid import RegularGrid
from ..crs import ProjectedCRS, GeographicalCRS, Cartesian
from . import _gdal
from . import _aai
from ..errors import GridError

def read_aai(fnm):
    """ Convenience function to open a ESRI ASCII grid and return a RegularGrid
    instance.
    """
    values, aschdr = _aai.aairead(fnm)
    t = {'xllcorner': aschdr['xllcorner'],
         'yllcorner': aschdr['yllcorner'],
         'dx'       : aschdr['cellsize'],
         'dy'       : aschdr['cellsize'],
         'sx'       : 0.0,
         'sy'       : 0.0}
    values[values==aschdr['nodata_value']] = np.nan
    return RegularGrid(t, values=values[::-1])

def proj4_isgeodetic(s):
    return ("lonlat" in s) or ("longlat" in s) or \
            ("latlon" in s) or ("latlong" in s)

def read_geotiff(fnm, in_memory=True, ibands=_gdal.ALL, **kw):
    """ Convenience function to open a GeoTIFF and return a RegularGrid
    instance.

    Parameters
    ----------
    fnm : str
        GeoTiff file path
    in_memory : bool
        if True (default), read entire dataset into memory
    ibands : int | list of ints, optional
        band(s) to open (default all)
    bandclass : Band class, optional
        class of band used by returned grid (default karta.band.CompressedBand)
        if in_memory is True, this parameter is ignored and the returned grid
        will have bands of type karta.raster._gdal.GdalFileBand
    """
    bands, hdr = _gdal.read(fnm, in_memory, ibands, **kw)

    t = {'xllcorner': hdr['xulcorner'] - hdr['ny'] * hdr['sx'],
         'yllcorner': hdr['yulcorner'] + hdr['ny'] * hdr['dy'],
         'dx'       : hdr['dx'],
         'dy'       : -hdr['dy'],
         'sx'       : hdr['sx'],
         'sy'       : -hdr['sy']}

    if len(hdr["srs"]["proj4"]) == 0:
        # invalid or missing SRS information
        crs = Cartesian
    elif proj4_isgeodetic(hdr["srs"]["proj4"]):
        geodstr = "+a={a} +f={f}".format(a=hdr["srs"]["semimajor"],
                                         f=hdr["srs"]["flattening"])
        crs = GeographicalCRS(geodstr, name=hdr["srs"]["name"])
    else:
        crs = ProjectedCRS(hdr["srs"]["proj4"], name=hdr["srs"]["name"])
    return RegularGrid(t, bands=bands, crs=crs, nodata_value=hdr["nodata"])

def from_geotiffs(*fnms, **kw):
    """ Read multiple GeoTiff files as bands within a single grid. Reads the
    first band within each GeoTiff, and checks that grids share the same
    spatial transform and coordinate system.

    Parameters
    ----------
    fnm : list of str
        GeoTiff file paths
    in_memory : bool
        if True (default), read entire dataset into memory
    bandclass : Band class, optional
        class of band used by returned grid (default karta.band.CompressedBand)
        if in_memory is True, this parameter is ignored and the returned grid
        will have bands of type karta.raster._gdal.GdalFileBand
    """
    in_memory = kw.get("in_memory", True)
    if len(fnms) == 0:
        raise ValueError("at least one filename must be provided")
    bands = []
    hdrs = []
    for i, fnm in enumerate(fnms):
        _b, _h = _gdal.read(fnm, in_memory, 1, **kw)
        if (i != 0) and (len(hdrs) != 0):
            for k, v in _h.items():
                if (isinstance(v, float)
                        and np.isnan(v) and not np.isnan(hdrs[-1].get(k, 0))):
                    if v != hdrs[-1].get(k, np.nan):
                        raise GridError("geotransform in {0} not equivalent to "
                                        "a previous GeoTiff".format(fnm))
        bands.append(_b[0])
        hdrs.append(_h)

    hdr = hdrs[0]
    t = {'xllcorner': hdr['xulcorner'] - hdr['ny'] * hdr['sx'],
         'yllcorner': hdr['yulcorner'] + hdr['ny'] * hdr['dy'],
         'dx'       : hdr['dx'],
         'dy'       : -hdr['dy'],
         'sx'       : hdr['sx'],
         'sy'       : -hdr['sy']}

    if proj4_isgeodetic(hdr["srs"]["proj4"]):
        geodstr = "+a={a} +f={f}".format(a=hdr["srs"]["semimajor"],
                                         f=hdr["srs"]["flattening"])
        crs = GeographicalCRS(geodstr, name=hdr["srs"]["name"])
    else:
        crs = ProjectedCRS(hdr["srs"]["proj4"], name=hdr["srs"]["name"])
    return RegularGrid(t, bands=bands, crs=crs, nodata_value=hdr["nodata"])

read_gtiff = read_geotiff



from libc.math cimport fmin, fmax, NAN
from vectorgeo cimport Vector2, fsign, bndlat_sph
from coordstring cimport CoordString

cpdef int crosses_dateline(double x0, double x1):
    """ check whether geodesic with longitudes x0, x1 [-180, 180) crosses the dateline.
    return -1 if crossing is west to east
    return 1 if crossing is east to west
    return 0 if no crossing
    """
    if (fsign(x0) != fsign(x1)) and (abs(x0-x1) > 180.0):
        if x1 - x0 > 180:
            return 1
        else:
            return -1
    else:
        return 0

def bbox(CoordString cs):
    """ Return a dateline-aware bbox for geographical coordinates. """
    cdef double xmin, ymin, xmax, ymax, segymin, segymax
    cdef double x0, y0, x1, y1
    cdef double rot
    cdef int i = 0, retint = 0, xdateline = 0, n = len(cs)

    if n == 0:
        return (NAN, NAN, NAN, NAN)

    if cs.ring:
        n += 1

    xmin = xmax = cs.getX(0)
    ymin = ymax = cs.getY(0)
    rot = 0
    x0 = cs.getX(0)
    y0 = cs.getY(0)

    while i != n:
        x1 = cs.getX(i)
        y1 = cs.getY(i)
        retint = bndlat_sph(Vector2(x0, y0), Vector2(x1, y1), &segymin, &segymax)
        if retint == 0:
            ymin = fmin(ymin, segymin)
            ymax = fmax(ymax, segymax)
        xdateline = crosses_dateline(x0, x1)
        if xdateline != 0:
            rot -= xdateline * 360.0
            xmin = fmin(xmin, x1+rot)
            xmax = fmax(xmax, x1+rot)
        else:
            if x0 > x1:
                xmin = fmin(xmin, x1)
            else:
                xmax = fmax(xmax, x1)
        x0 = x1
        y0 = y1
        i += 1

    xmin = (xmin+180) % 360 - 180
    xmax = (xmax+180) % 360 - 180
    return (xmin, ymin, xmax, ymax)
